/*
 * isolconsole.c
 *      IPMI Serial-Over-LAN console application
 *
 * Author:  Andy Cress  arcress@users.sourceforge.net
 * Copyright (c) 2006 Intel Corporation. 
 *
 * 10/12/06 Andy Cress - created, just a framework for now
 * 02/08/07 Andy Cress - now working for IPMI LAN 2.0 SOL in Linux,
 *                       still stdio problems in Windows
 * 02/23/07 Andy Cress - fixed Windows stdin, still Windows output issues,
 *                       need more in xlate_vt100_win().
 * 04/06/07 Andy Cress - more vt100-to-windows translation, 
 *                       added -r for raw termio
 * 04/19/07 Andy Cress - more vt100-to-windows translation, 
 * 05/08/07 Andy Cress - more logic for 1.5 SOL.  Session opens but 
 *                       1.5 SOL data packets are not yet working.
 * 05/24/07 Andy Cress - Fix Enter key confusion in BIOS Setup (use CR+LF)
 * 06/15/07 Andy Cress - More fixes to Windows translation (fUseWinCon)
 *
 * NOTE: There are a few other options for Serial-Over-LAN console 
 * applications:
 *   ipmitool.sf.net (has v2 sol console capability for Linux)
 *   Intel dpccli (CLI console for Linux and Windows)
 *   Intel ISM console (for Windows)
 *   Intel System Management Software (LANDesk) console 
 * The Intel applications support both IPMI 1.5 and 2.0 SOL 
 * protocols.
 */
/*M*
Copyright (c) 2006, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

  a.. Redistributions of source code must retain the above copyright notice, 
      this list of conditions and the following disclaimer. 
  b.. Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation 
      and/or other materials provided with the distribution. 
  c.. Neither the name of Intel Corporation nor the names of its contributors 
      may be used to endorse or promote products derived from this software 
      without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR 
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *M*/
#ifdef WIN32
#include <windows.h>
#include <stdio.h>
#include <winsock.h>
#include <io.h>
#include <conio.h>
#include <wincon.h>
#include "getopt.h"
#else
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <getopt.h>
#endif
#include <string.h>
#include <time.h>
#include "ipmicmd.h"
#include "ipmilanplus.h"
 
// #define DBG  1
#define ACTIVATE_PAYLOAD    0x48
#define DEACTIVATE_PAYLOAD  0x49
#define PAYLOAD_TYPE_SOL    0x01
#define SOL_SERIAL_ALERT_MASK_DEFERRED 0x04
#define SOL_BMC_ASSERTS_CTS_MASK_TRUE  0x00  
#define SOL_BMC_ASSERTS_CTS_MASK_FALSE 0x02
#define AUTHTYPE_RMCP_PLUS  0x06  
#define RV_END   -2
#define CH_CR   '\r'    /*0x0D =='\r'*/
#define CH_LF   '\n'    /*0x0A =='\n'*/
#define CH_ESC  '\033'  /*0x1B == ASCII Escape */
#define CH_CTL  '\316'  /*0xCE == VT100 Control */
#define CH_DEL  '\177'  /*0x7F == Delete key */
#define CH_BS   '\010'  /*0x08 == Backspace */
#define CH_TAB  '\011'  /*0x09 == Tab */

typedef struct {
  int type;
  int len;
  char *data;
  } SOL_RSP_PKT;

extern void tty_setraw(int mode);      /*from ipmicmd.c*/
extern void tty_setnormal(int mode);   /*from ipmicmd.c*/
extern int  lan_get_fd(void);           /*from ipmilan.c*/
extern int  lan_send_sol( uchar *payload, int len, SOL_RSP_PKT *rsp);
extern int  lan_recv_sol( SOL_RSP_PKT *rsp );
extern int  lan_keepalive(int type);
extern void lan_get_sol_data(uchar fEnc, uchar iseed, uint32 *seed);
extern void lan_set_sol_data(uchar fEnc, uchar auth, uchar iseed, 
				int insize, int outsize);
#ifdef HAVE_LANPLUS
extern int  lan2_get_fd(void);          /*from ipmilanplus.c*/
extern int  lan2_send_sol( uchar *payload, int len, SOL_RSP_PKT *rsp);
extern int  lan2_recv_sol( SOL_RSP_PKT *rsp );
extern void lan2_set_sol_data(int insize, int outsize, int port, void*hnd);
extern int  lan2_keepalive(int type);
#else
int  lan2_get_fd(void) { return(-1); }
int  lan2_send_sol( uchar *payload, int len, SOL_RSP_PKT *rsp) { return(-1); }
int  lan2_recv_sol( SOL_RSP_PKT *rsp ) { return(-1); }
void lan2_set_sol_data(int insize, int outsize, int port, void*hnd)
{ return; }
int  lan2_keepalive(int type) { return(-1); }
#endif

/*
 * Global variables 
 */
static char * progver   = "1.6";
static char * progname  = "isolconsole";
static char  fdebug        = 0;
static char  fpicmg        = 0;
static char  factivate     = 0;
static char  fdeactivate   = 0;
static char  fUseWinCon    = 0;
static char  bDriver       = 0;
static uchar bSolVer       = 0;
static uchar bKeepAlive    = 2;
static char  sol_esc_seq[3] = {'~',0,0};  /* SOL escape char ('~') */
static uchar ipmi_maj = 0;
static uchar ipmi_min = 0;
static int   mfg      = 0;  /* Manufacturer IANA vendor id */
static uchar fRaw     = 0;
static uchar fCRLF = 0;     /* =1 to use legacy CR/LF translation for BIOS */
static uchar bTextMode = 0;  
static uchar fSolEncryption     = 1;
static uchar fSolAuthentication = 1;
static int done = 0;

#ifdef WIN32
/*======== start Windows =============================*/
    /* Windows os_read, os_select and supporting code */
#define OUTBUF_SZ   128
#define INBUF_SZ    64
static uchar rg_stdin[INBUF_SZ];
static int   n_stdin = 0;
static HANDLE g_hConOut = INVALID_HANDLE_VALUE;
static SMALL_RECT g_srScrollRegion;
static COORD      g_coord;
static int g_Rows = 25;
static int g_Columns = 80;
/* extern int _cdecl _getche(void);   * see conio.h */
/* extern int _cdecl _getch(void);    * see conio.h */

#define  NESCS   50
int esc_type = 0;  
static struct { 
   int  vt_sz;
   char vt[16];
   char win[16];
   int  win_sz;
} xlate_vt[NESCS] = {
/* 0*/  3, {CH_ESC,"[K"},   { "" }, 0,  /*erase line */
/* 1*/  3, {CH_ESC,"[J"},   { "" }, 0,  /*erase screen*/
/* 2*/ 11, {CH_ESC,"[00;00m00m"}, {CH_ESC,"|00;00m" },8, /*v;h cursor position*/
/* 3*/  8, {CH_ESC,"[00;00m"},    {CH_ESC,"|00;00m" },8, /*v;h cursor position*/
/* 4*/  5, {CH_ESC,"[00m"},    { CH_ESC,"|00m" }, 0,    /*0=normal text mode*/
/* 5*/  6, {CH_ESC,"[0;0H"},   {CH_ESC,"|0;0H" }, 6,    /*v;h home position*/ 
/* 6*/  7, {CH_ESC,"[00;0H"},  {CH_ESC,"|00;0H" }, 7,   /*v;h home position*/
/* 7*/  7, {CH_ESC,"[00;0H"},  {CH_ESC,"|00;0H" }, 7,   /*v;h home position*/
/* 8*/  8, {CH_ESC,"[0;000H"}, {CH_ESC,"|0;000H" }, 8,  /*v;h home position*/
/* 9*/  8, {CH_ESC,"[00;00H"}, {CH_ESC,"|00;00H" }, 8,  /*v;h home position*/
/*10*/  3, {CH_ESC,"[H"},      { "" }, 0,    /* home */
/*11*/  8, {CH_ESC,"[00:00r"}, { "" }, 0,  
/*12*/  7, {CH_ESC,"[0:00r"},  { "" }, 0, 
/*13*/  5, {CH_ESC,"[00A"}, {CH_CTL,"H"}, 2,      /*cursor up  */
/*14*/  5, {CH_ESC,"[00B"}, {CH_CTL,"P"}, 2,      /*cursor down  */
/*15*/  5, {CH_ESC,"[00C"}, {CH_CTL,"M" }, 2,     /*cursor right  */ 
/*16*/  5, {CH_ESC,"[00D"}, {CH_CTL,"K" }, 2,     /*cursor left  */
/*17*/  3, {CH_ESC,"[A"},   {CH_CTL,"H"}, 2,      /*cursor up  */
/*18*/  3, {CH_ESC,"[B"},   {CH_CTL,"P"}, 2,      /*cursor down  */
/*19*/  3, {CH_ESC,"[C"},   {CH_CTL,"M"}, 2,      /*cursor right  */
/*20*/  3, {CH_ESC,"[D"},   {CH_CTL,"K"}, 2,      /*cursor left  */
/*21*/  3, {CH_ESC,"OA"},   {CH_CTL,"H"}, 2,      /*cursor up  */
/*22*/  3, {CH_ESC,"OB"},   {CH_CTL,"P"}, 2,      /*cursor down  */
/*23*/  3, {CH_ESC,"OC"},   {CH_CTL,"M"}, 2,      /*cursor right  */
/*24*/  3, {CH_ESC,"OD"},   {CH_CTL,"K"}, 2,      /*cursor left  */
/*25*/  6, {CH_ESC,"[0;0f"}, {CH_ESC,"|0;0f" }, 6,  /*cursor position*/
/*26*/  7, {CH_ESC,"[00;0f"}, {CH_ESC,"|00;0f" }, 7,  /*cursor position*/
/*27*/  7, {CH_ESC,"[0;00f"}, {CH_ESC,"|0;00f" }, 7,  /*cursor position*/
/*28*/  8, {CH_ESC,"[00;00f"}, {CH_ESC,"|00;00f" }, 8,  /*cursor position*/
/*29*/  4, {CH_ESC,"[0A"}, {CH_CTL,"H"}, 2,      /*cursor up  */
/*30*/  4, {CH_ESC,"[0B"}, {CH_CTL,"P"}, 2,      /*cursor down  */
/*31*/  4, {CH_ESC,"[0C"}, {CH_CTL,"M" }, 2,     /*cursor right  */ 
/*32*/  4, {CH_ESC,"[0D"}, {CH_CTL,"K" }, 2,     /*cursor left  */
/*33*/  5, {CH_ESC,"[Pm0"}, { "" }, 0,  /*char attributes*/
/*34*/  2, {CH_ESC,"E"},    { "" }, 0,  /*next line */
/*35*/  2, {CH_ESC,"="},    { "" }, 0,  
/*36*/  5, {CH_ESC,"[?0h"}, { "" }, 0,  
/*37*/  5, {CH_ESC,"[?0l"}, { "" }, 0,  
/*38*/  6, {CH_ESC,"[?00h"}, { "" }, 0,  
/*39*/  6, {CH_ESC,"[?00l"}, { "" }, 0,  
/*40*/  3, {CH_ESC,"0h"}, { "" }, 0,  
/*41*/  3, {CH_ESC,"0l"}, { "" }, 0,  
/*42*/  4, {CH_ESC,"00h"}, { "" }, 0,  
/*43*/  4, {CH_ESC,"00l"}, { "" }, 0,  
/*44*/  3, {CH_ESC,"[m"},  { "" }, 0, /*0=normal text mode*/
/*45*/  4, {CH_ESC,"[0m"}, { "" }, 0, /*text mode 0=normal 1=BIOS*/
/*46*/  4, {CH_ESC,"[0J"}, { "" } , 0,  /*clear screen, etc */
/*47*/  4, {CH_ESC,"[0K"}, { "" } , 0,  /*clear line, etc */
/*48*/  3, {CH_ESC,"(0"},   { "" }, 0, /*character set*/
/*49*/  2, {CH_ESC,">"},    { "" }, 0  
};


static int EscMatch(uchar *p1, uchar *p2, int len)
{
   int i;
   int rv = 0;
   for (i = 0; i < len; i++) {
      if (p1[i] == p2[i]) continue;
      else if ((p1[i] >= '0') && (p1[i] <= '9') && 
               (p2[i] >= '0') && (p2[i] <= '9')) continue;
      else break;
   }
   if (i >= len) rv = 1;
   return(rv);
}

#define NCTLS  5
static int IsCtl(uchar c)
{
   char ctls[NCTLS] = {CH_LF, CH_CR, CH_DEL, CH_BS, CH_TAB};
   int i;
   for (i = 0; i < NCTLS; i++)
      if (c == ctls[i]) break;
   if (i < NCTLS) return(1);
   else return(0);
}

static FILE *fplog = NULL;
void dbglog( char *pattn, ... )
{
    va_list arglist;
    if (fplog == NULL) fplog = fopen("isol.log","a");
    if (fplog == NULL) return;
    va_start( arglist, pattn );
    vfprintf( fplog, pattn, arglist );
    va_end( arglist );
    fclose(fplog);
    fplog = NULL;
}

static int IsVtEsc(uchar c, uchar *buf, int idx)
{
   int ret = 0;
   uchar tmpbuf[20];
   int i;

   /* add this char to the esc sequence for matching */
   memcpy(tmpbuf,buf,idx);
   tmpbuf[idx] = c;
   idx++;
   tmpbuf[idx] = 0;  /*stringify*/
   /* see if it matches a known VT esc sequence */
   for (i = 0; i < NESCS; i++) {
      if (idx == 0) {
          if (c == CH_ESC) ret = 1; 
          break; 
      } else if (EscMatch(xlate_vt[i].vt,tmpbuf,idx)) { /*matches*/
          ret = 1;
          if (idx >= xlate_vt[i].vt_sz) {
             esc_type = i;
             ret = 2; /*complete*/
             if (fdebug) {
	       printf("esc match, type[%d]: %s\n",esc_type,&tmpbuf[1]);
	       dbglog("esc match, type[%d]: %s\n",esc_type,&tmpbuf[1]);
             }
             break;
          }
      } 
#ifdef OLD
      else {  /*not a match*/
          if (idx > 1) {   /* have some chars in escbuf already */
             tmpbuf[idx]   = 0;  /*stringify*/
          }
      }
#endif
   }  /*end NESCS loop*/
   if (fdebug && (idx > 1) && (ret == 0)) {
	printf("no match, esc seq(%d): %s, c=%c\n",idx,&tmpbuf[1],c);
	dbglog("no match, esc seq(%d): %s, c=%c\n",idx,&tmpbuf[1],c);
   }
   return(ret);
}

static void console_write_chars(uchar *buf, int sbuf)
{
   DWORD written;
   uchar c;
   int i;

   if (g_hConOut != INVALID_HANDLE_VALUE) {
      /* When writing fails or didn't write a single character, pretend 
       * one character was written, otherwise we get stuck. */
      /* Normally Unicode, but WriteConsoleOutputCharacterA is ANSI */
      if (WriteConsoleOutputCharacterA(g_hConOut,buf,sbuf,g_coord,&written) == 0 
          || written == 0)
             written = 1;
#ifdef DBG
      buf[sbuf] = 0;  /*stringify*/
      dbglog("x = %d + %d(%d), buf: %s\n",g_coord.X,sbuf,written,buf);
#endif
      g_coord.X += sbuf;
      SetConsoleCursorPosition(g_hConOut, g_coord);
   } else {  /*do stdout method*/
      for (i = 0; i < sbuf; i++) {
          c = buf[i];
          /* do special CH_CR & CH_BS handling here? */
          putc(c, stdout);
      }
      fflush(stdout);
   }
}

static void console_get_coord(void)
{
   CONSOLE_SCREEN_BUFFER_INFO csbi;
   /* leave unchanged if any errors */
   if (g_hConOut != INVALID_HANDLE_VALUE) {
      if (GetConsoleScreenBufferInfo(g_hConOut, &csbi)) {
          g_coord.X = csbi.dwCursorPosition.X;
          g_coord.Y = csbi.dwCursorPosition.Y;
      }
#ifdef DBG
      dbglog("console_get_coord: (%d,%d) max=(%d,%d)\n", 
			g_coord.X, g_coord.Y , g_Columns, g_Rows );
#endif
   }
}

static int goto_xy( unsigned x, unsigned y)
{
    if (g_hConOut == INVALID_HANDLE_VALUE) return(-1);
    if (fdebug) printf("goto_xy(%d,%d)\n",x,y);
#ifdef DBG
    dbglog("goto_xy(%d,%d)\n",x,y);
#endif
    if (x < 0 || x >= (unsigned)g_Columns || y < 0 ) 
                                       // || y >= (unsigned)g_Rows)
        return(-1);
    /* zero-based coordinates */
    g_coord.X = x;
    g_coord.Y = y;
    SetConsoleCursorPosition(g_hConOut, g_coord);
    return(0);
}

static int goto_x( int x )
{
    if (g_hConOut == INVALID_HANDLE_VALUE) return(-1);
    g_coord.X = x;
    // same g_coord.Y 
#ifdef DBG
    dbglog("goto_x(%d,%d)\n",x,g_coord.Y);
#endif
    if (fdebug) printf("goto_x(%d,%d)\n", g_coord.X, g_coord.Y);
    SetConsoleCursorPosition(g_hConOut, g_coord);
    return(0);
}

static int goto_nextline( void )
{
    if (g_hConOut == INVALID_HANDLE_VALUE) return(-1);
    /* zero-based coordinates */
    g_coord.X = 0;
    g_coord.Y += 1;
    if (fdebug) printf("goto_nextline(%d,%d)\n", g_coord.X, g_coord.Y);
#ifdef DBG
    dbglog("goto_nextline(%d,%d)\n", g_coord.X, g_coord.Y);
#endif
    SetConsoleCursorPosition(g_hConOut, g_coord);
    return(0);
}

static int console_backspace( void )
{
    DWORD dwDummy;
    if (g_hConOut == INVALID_HANDLE_VALUE) return(-1);
    if (g_coord.X > 0) {
       g_coord.X--;
       FillConsoleOutputCharacterA(g_hConOut,' ',1,g_coord, &dwDummy);
       SetConsoleCursorPosition(g_hConOut, g_coord);
    }
    return(0);
}

static int clear_line( void)
{
    DWORD dwDummy;
    unsigned int n;
    if (g_hConOut == INVALID_HANDLE_VALUE) return(-1);
    else {
       n = g_Columns - g_coord.X;
       FillConsoleOutputCharacterA(g_hConOut,' ',n,g_coord, &dwDummy);
    } 
    return(0);
}

static int clear_screen( void)
{
    DWORD dwDummy;
    unsigned int n;
    if (g_hConOut == INVALID_HANDLE_VALUE) return(-1);
    else {
       g_coord.X = 0;
       g_coord.Y = 0;
       SetConsoleCursorPosition(g_hConOut, g_coord);
       n = g_Columns * g_Rows;
       FillConsoleOutputCharacterA(g_hConOut,' ',n,g_coord, &dwDummy);
       g_coord.X = 0;
       g_coord.Y = 0;
       SetConsoleCursorPosition(g_hConOut, g_coord);
    } 
    return(0);
}

static void parse_xy(uchar *buf, int sbuf, int *x, int *y)
{
    int r,c;    /*row, col*/
    char *p;
    /* Input buf is a VT100 esc sequence, like {ESC,"[00:00H"} 
     * So, r comes from buf[2,3], and c is usu buf[5,6].  */
    c = 0;
    r = 0;
    p = strchr(buf,';');
    buf[sbuf-1] = 0;  /*end at H*/
    if (p != NULL) { 
	p[0] = 0; 
        r = atoi(&buf[2]);
	p++; 
	c = atoi(p);
    }
    else c = atoi(&buf[2]);
    *x = c;
    *y = r;
}

static void console_write_esc(int itype, uchar *ebuf, int elen, 
				uchar *obuf, int *olen)
{
     int n, len; 
     int x,y;

     /* For this vt esc, output valid ANSI */
     n = *olen;
     len = xlate_vt[itype].win_sz;
     switch (itype) {
           case 0:   /* "[K" clear line */
           case 47:  /* "[2K" clear line */
#ifdef DBG
                dbglog("%d: clear_line from (%d,%d)\n", 
				itype,g_coord.X, g_coord.Y);
#endif
		if (clear_line() != 0) 
                    obuf[n++] = '\b';  /*0x08,backspace*/
                break;
           case 1:   /* "[J" clear screen*/
           case 46:  /* "[2J" clear screen */
		if (clear_screen() != 0) {
		   obuf[n++] = '\f';  /* 0x0c, FormFeed*/
                }
                break;
           case 2:
           case 3:  /*"[00:00m" text mode */
                parse_xy(ebuf,elen,&x,&y);
#ifdef DBG
                dbglog("%d: [00m esc (%d,%d)\n", itype,x,y);
#endif
		/* do not goto_x(x) */
                break;
           case 4:  /* "[00m" text mode change */
           case 45:  /* "[0m" text mode change */
                parse_xy(ebuf,elen,&x,&y);
                if (x == 0) bTextMode = 0;
                if (x == 1) bTextMode = 1;
                if (fdebug) printf("Text Mode changed to %d\n",bTextMode);
                break;
           case 5:  /*cursor positioning "[x;yH"*/
           case 6:  /*cursor positioning "[x;yH"*/
           case 7:  /*cursor positioning "[x;yH"*/
           case 8:  /*cursor positioning "[x;yH"*/
           case 9:  /*cursor positioning "[x;yH"*/
                parse_xy(ebuf,elen,&x,&y);
		if (goto_xy(x,y) != 0) {
                   obuf[n++] = '\n';  /* 0x0a, LineFeed*/
                }
                break;
           case 25:  /* "[x;yf"*/
           case 26:  /* "[x;yf"*/
           case 27:  /* "[x;yf"*/
           case 28:  /* "[x;yf"*/
                parse_xy(ebuf,elen,&x,&y);
		if (goto_xy(x,y) != 0) {
                   obuf[n++] = '\b';  /* 0x08, backspace */
                }
                break;
           case 44:  /* "[m" normal text mode */
                break;
	   /*TODO: add more here */
           default:
#ifdef DBG
                dbglog("%d: default, no action\n", itype);
#endif
                ; /*do nothing by default*/
     }
     *olen = n;
}

static void xlate_vt100_win(uchar *pin, int ilen, uchar *pout, int *olen)
{
    int i, j, k, l;
    static int iesc = 0;
    static uchar escbuf[16];
    static uchar outtmp[16];
    uchar *pdata;

    /* Translate vt100 escape sequences to ansi here. */
    j = 0;   /*index into pout*/
    l = 0;   /*length of pdata*/
    pdata = pout;
    for (i = 0; i < ilen; i++) {
       if (j >= OUTBUF_SZ) break;
       k = IsVtEsc(pin[i],escbuf,iesc);
       if (k > 0) {  /*vt100 esc sequence*/
          escbuf[iesc++] = pin[i];
          if (k == 2) {  /*sequence complete, flush and start over*/
             escbuf[iesc]   = 0;  /*stringify*/
             if (fdebug) printf("xlate esc seq(%d): %s\n",iesc,&escbuf[1]);
             console_write_chars(pdata,l);  /*write what we have so far*/
             l = 0;
             pdata = &pout[j];
             console_write_esc(esc_type,escbuf,iesc,outtmp,&l); /*escape seq*/
             iesc = 0;
             j += l;
          }
       } else {  /*standard chars pass through*/
          iesc = 0;
          if ((fUseWinCon) && IsCtl(pin[i])) {
             console_write_chars(pdata,l);  /*write what we have so far*/
             pdata = &pout[j];
             l = 0;
             switch(pin[i]) {
               case CH_CR: // goto_x(0);  break;
		 if (fCRLF) break;  /*wait for LF*/
               case CH_LF: goto_nextline();  break;
               case CH_BS: console_backspace();  break;
               case CH_DEL:  /* same as BS? */
               case CH_TAB:  
	       default:   
		        dbglog("skipped ctl %02x\n",pin[i]); /*++++*/
			break;
             }
          } else {
             pout[j++] = pin[i];
             l++;
          }
       }
    }
    console_write_chars(pdata,l);  /*write the rest*/
    *olen = j;
}

static void console_open(void)
{
   if (fUseWinCon && !fRaw) 
      g_hConOut = GetStdHandle(STD_OUTPUT_HANDLE); 
   g_coord.X = 0;
   g_coord.Y = 0;  
   console_get_coord();
}

static void console_close(void)
{
   if (g_hConOut != INVALID_HANDLE_VALUE) {
      CloseHandle(g_hConOut);
      g_hConOut = INVALID_HANDLE_VALUE;
   }
}

DWORD WINAPI input_thread( LPVOID lpParam) 
{
   // HANDLE hstdin;
   DWORD rv = 0;
   int c;

   // hstdin = GetStdHandle(STD_INPUT_HANDLE); 
   while (!done)
   {
      c = _getch();   /*was getc(stdin) or _getche() */
      if (c != EOF) {
         if (n_stdin < sizeof(rg_stdin)) {
            rg_stdin[n_stdin] = c;
            n_stdin++;
            // if (fdebug) printf("getche() got 0x%02x, %d\n",c,n_stdin);
         } 
         // else if (fdebug) printf("getche() n_stdin=%d > max\n",n_stdin);
      } 
      // else if (fdebug) printf("getche() == EOF\n");
   }
   return(rv);
}

static int os_read(int fd, uchar *buf, int sz)
{
    int len = 0;
    int i;
    if ((fd == 0) && (n_stdin > 0)) {
       for (i = 0; i < n_stdin; i++)
          buf[i] = rg_stdin[i];
       len = n_stdin;
       n_stdin = 0;
    } else {
       // ReadFile(fd, buf, sz, &len, NULL);
       // len = _read(fd, buf, sz);
       len = recv(fd, buf, sz, 0);
    }
    return(len);
}

int os_select(int infd, int sfd, fd_set *read_fds)
{
   int rv = 0;
   struct timeval tv;
   fd_set err_fds;
   int rv0 = -1;

   /* check the socket for new data via select */
   /* Windows select only works on WSA sockets */
   {
      FD_ZERO(read_fds);
      FD_SET(sfd, read_fds);
      FD_ZERO(&err_fds);
      tv.tv_sec =  0;
      tv.tv_usec = 50000; /* 50 ms */
      rv = select(sfd + 1, read_fds, NULL, NULL, &tv);
      if (rv < 0) {
         rv = -(WSAGetLastError());
         if (fdebug) printf("select(%d) error %d\n",sfd,rv);
      }
      else if (FD_ISSET(sfd, &err_fds)) {
         if (fdebug) printf("select(%d) err_fds rv=%d err=%d\n",sfd, rv, 
			 WSAGetLastError());
      }
      if (rv > 0 && fdebug) printf("select(%d) got socket data %d\n",sfd,rv);
   }  
   if (infd == 0) {  /* check stdin */
      rv0 = -1;
      if (n_stdin > 0) rv0 = 0;
   }
   if (rv0 == 0) {
      FD_SET(infd, read_fds);
      if (rv < 0) rv = 1;
      else rv++;
   }
   return rv;
}
/*======== end Windows ===============================*/
#else
/*======== start Linux ===============================*/
static int os_read(int fd, uchar *buf, int sz)
{
    int len = 0;
    len = read(fd, buf, sz);
    return(len);
}

int os_select(int infd, int sfd, fd_set *read_fds)
{
   int rv = 0;
   struct timeval tv;

   /* Linux handles both stdin & socket via select() */
   FD_ZERO(read_fds);
   FD_SET(infd, read_fds);
   FD_SET(sfd, read_fds);
   tv.tv_sec =  0;
   tv.tv_usec = 500000;
   rv = select(sfd + 1, read_fds, NULL, NULL, &tv);
   return rv;
}

void CheckTextMode(uchar *buffer, int sbuf)
{
    int c, i, j, imatch;
    uchar pattern[4] = {CH_ESC,'[','0','m'};
    int spattern = 4;
 
    /* if CH_ESC,"[0m", TextMode change */
    j = 0;
    imatch = 0;
    for (j = 0; j < sbuf; j++) {
        if ((sbuf - j) < spattern && imatch == 0) break;
        c = buffer[j];
        if (c == pattern[imatch]) {
            imatch++;
        } else if (pattern[imatch] == '?') {  /*wildcard char*/
            imatch++;
        } else if (pattern[imatch] == '0' && (c>='0' && c <='9')) { /*numeric*/
            imatch++;
        } else {
            if (imatch > 0) {
               if (j > 0) j--; /* try again with the first match char */
               imatch = 0;
            }
        }
        if (imatch == spattern) { /*match found*/
            i = (j+1) - imatch;  /*buffer[i] is the match */
            bTextMode = buffer[i+2] & 0x0f;  /* usu 0 or 1 */
            if (fdebug) printf("Text Mode changed to %d\n",bTextMode);
        }
    }
    return;
}

/*======== end Linux =================================*/
#endif

static int sol_send(uchar *buf, int len, SOL_RSP_PKT *rsp)
{
    int rv = 0;
    if (bSolVer == 1) {
       rv = lan_send_sol(buf, len, rsp);
    } else 
       rv = lan2_send_sol(buf, len, rsp);
    return(rv);
}

static int sol_recv( SOL_RSP_PKT *rsp)
{
    int rv = 0;
    if (bSolVer == 1) {
       rv = lan_recv_sol( rsp );
    } else 
       rv = lan2_recv_sol( rsp );
    return(rv);
}

static int  sol_get_fd(void)
{
    int fd = 0;
    if (bSolVer == 1) {
       fd = lan_get_fd();
    } else 
       fd = lan2_get_fd();
    return(fd);
}

#define SOL_TIMEOUT   30
static int sol_keepalive(int type)
{
   int rv = 0;
   static time_t keepalive_start; 
   time_t end; 

   time(&end); 
   if (end - keepalive_start > SOL_TIMEOUT) {
       if (bSolVer == 1) {
           rv = lan_keepalive(bKeepAlive);   
       } else 
           rv = lan2_keepalive(bKeepAlive);   
       time(&keepalive_start);
       if (fdebug) printf("sol_keepalive sent, rv = %d\n",rv);
   }
   return(rv);
}

void sol_output_handler(SOL_RSP_PKT *rsp)
{
    uchar *pdata;
    int len;
    if (rsp &&
        (rsp->type == PAYLOAD_TYPE_SOL))
    {
#ifdef WIN32
         uchar buf[OUTBUF_SZ];

	 if (fRaw) {
            pdata = rsp->data;
            len   = rsp->len;
            console_write_chars(pdata,len);
         } else {
            pdata = buf;
            xlate_vt100_win(rsp->data,rsp->len,pdata,&len);
         }
         if (fdebug) printf("sol_output: len = %d\n",len);
#else
         int i;
         int mode;
         uchar c;

         pdata = rsp->data;
         len   = rsp->len;
         if (fdebug) {
             printf("sol_output: len = %d\n",len);
             dump_log("sol_output",pdata,len,1); /*like printlog*/
         }
         CheckTextMode(pdata,len);
         for (i = 0; i < len; i++) {
	      c = pdata[i];
              /* do special CH_CR handling here? */
              putc(c, stdout);
         }
         fflush(stdout);
#endif
    }
}

static int    sol_retries = 3;
static uchar  sol15_iseed = 0x01;  /*initial seed count, usu 0 or 1 */
static uint32 sol15_wseed = 0;     /*32-bit seed value*/

int sol_input_handler( uchar *input, int ilen)
{
   int rv = 0;
   uchar payload[0x64]; 
   SOL_RSP_PKT  rs;
   int i, length, t;
   uchar c, lastc;

   memset(&payload, 0, sizeof(payload)); 
   length = 0;
   rs.len = 0;
   for (i = 0; i < ilen; ++i)
   {
        c = input[i];

        /* look for escape sequences */
        // if (lastc == '\n' && c == sol_esc_ch) 
        if (c == sol_esc_seq[0]) {  /*'~'*/
           return(RV_END);
        }
        if (fdebug && (c == CH_CR || c == CH_LF))
	    printf("sol_send: i = %d, c = %02x\n", i,c);
        /*
         * Note that this CR/LF logic works for BIOS, but causes
         * extra returns after Linux commands. 
         * TODO: fix to set fCRLF from bTextMode
         */
#ifdef WIN32
        if (fUseWinCon && !fRaw) {
           if (c == CH_CR) c = CH_LF;
        } else {
           if (c == CH_CR) { /*0x0D = '\r' from Windows Enter*/
              if (fCRLF) payload[length++] = c;
              if (!fRaw) c = CH_LF;    /*0x0A = '\n' newline */
           }
           // if (!fRaw && (c == CH_DEL)) { /* DEL & Backspace translation */
           //    payload[length++] = CH_BS;
           // }
        }
#else
        if (fCRLF && c == CH_LF) { /*0x0A = '\n' from Linux Enter*/
            payload[length++] = CH_CR;
            /* add c==CH_LF below */
        }
#endif

        /* copy other data to payload */
        payload[length++] = c;
        lastc = c;
   }
   if (length > 0) {
      for (t = 0; t < sol_retries; t++) {
         rv = sol_send(payload, length, &rs);
         if (fdebug) 
		printf("sol_send(%d,%d) rv = %d (0x%02x)\n", t,length,rv,rv);
         if (rv == 0) break;
         os_usleep(0,5000);
      }
      if (rv >= 0 && rs.len != 0) { /*handle the rsp data*/
          if (fdebug) printf("output: handler(%d)\n",rs.len);
          sol_output_handler(&rs);
      }
   }
   return(rv);
}

static int send_activate_15sol(void)
{
   int rv = 0;
   uchar actcmd;
   uchar netfn;
   uchar ibuf[64];
   uchar rbuf[64];
   int ilen, rlen;
   uchar cc;

   lan_get_sol_data(fSolEncryption,sol15_iseed,&sol15_wseed);
   actcmd  = (uchar)(ACTIVATE_SOL1 & 0x00ff); 
   netfn   = NETFN_SOL; 
   if (fSolEncryption) ibuf[0] = 0x11;   /* use encryption, activate SOL */
   else ibuf[0] = 0x01;   /*no encryptionn, no serial alerts, activate SOL */
   memcpy(&ibuf[1],&sol15_wseed,4);  /*32-bit seed*/
   ilen = 5;
   rlen = sizeof(rbuf);
   rv = ipmi_cmdraw(actcmd,netfn,BMC_SA, PUBLIC_BUS, BMC_LUN,
                   ibuf,(uchar)ilen,rbuf,&rlen,&cc,fdebug);
   if (rv == 0 && cc != 0) rv = cc;
   /* switch(cc) { case 0x80: SOL already active on another session
    *              case 0x81: SOL disabled
    *              case 0x82: SOL activation limit reached  }
    */
   if (rv == 0) {  /*success*/
      if (fdebug) dump_buf("sol 15 act_resp",rbuf,rlen,0);
      /* sol 15 act_resp (len=4): 0000: 00 01 64 64 */
      /* 00=auth, 01=seed_cnt, 64=in_payload_sz, 64=out_payload_sz */
      lan_set_sol_data(fSolEncryption,rbuf[0],rbuf[1], rbuf[2], rbuf[3]);
   }
   return(rv);
}

static int send_deactivate_15sol(void)
{
   int rv = -1;

   uchar actcmd;
   uchar netfn;
   uchar ibuf[64];
   uchar rbuf[64];
   int ilen, rlen;
   uchar cc;

   /* SOL for IPMI 1.5 does not have an explicit deactivate, so
    * use an activate with data0 = 00. 
    */
   actcmd  = (uchar)(ACTIVATE_SOL1 & 0x00ff); 
   netfn   = NETFN_SOL; 
   if (fSolEncryption) ibuf[0] = 0x10;  /* deactivate, use encryption */
   else ibuf[0] = 0x00;    /*deactivate*/
   memcpy(&ibuf[1],&sol15_wseed,4);  /*32-bit seed*/
   ilen = 5;
   rlen = sizeof(rbuf);
   rv = ipmi_cmdraw(actcmd,netfn,BMC_SA, PUBLIC_BUS, BMC_LUN,
                   ibuf,(uchar)ilen,rbuf,&rlen,&cc,fdebug);
   if (rv == 0 && cc != 0) rv = cc;
   if (fdebug) dump_buf("sol 15 deact_resp",rbuf,rlen,0); /* 80 00 32 ff */
   return(rv);
}

int send_activate_sol(uchar verSOL)
{
   int rv = 0;
   int in_payload_size, out_payload_size, udp_port;
   uchar actcmd;
   uchar netfn;
   uchar ibuf[64];
   uchar rbuf[64];
   int ilen, rlen;
   uchar cc;
   /* activate SOL = 0x01, get SOL status = 0x05 */
   if (verSOL == 1) {
      rv = send_activate_15sol();
      return(rv);
   } else {   /* use IPMI 2.0 method */
      actcmd  = ACTIVATE_PAYLOAD;
      netfn   = NETFN_APP;  
      ibuf[0] = PAYLOAD_TYPE_SOL;
      ibuf[1] = 1;        /* payload instance */
      ibuf[2] = SOL_SERIAL_ALERT_MASK_DEFERRED;
      if (fSolEncryption)     ibuf[2] |= 0x80;
      if (fSolAuthentication) ibuf[2] |= 0x40;
      if (mfg == VENDOR_INTEL) 
           ibuf[2] |= SOL_BMC_ASSERTS_CTS_MASK_TRUE;
      else ibuf[2] |= SOL_BMC_ASSERTS_CTS_MASK_FALSE;
      ibuf[3] = 0;
      ibuf[4] = 0;
      ibuf[5] = 0;
      ilen = 6;
   }
   rlen = sizeof(rbuf);
   rv = ipmi_cmdraw(actcmd,netfn,BMC_SA, PUBLIC_BUS, BMC_LUN,
                   ibuf,(uchar)ilen,rbuf,&rlen,&cc,fdebug);
   if (fdebug) {
      dump_buf("sol act_cmd",ibuf,ilen,0);
      printf("send_activate v2(%x,%x) rv = %d cc = %x\n",actcmd,netfn,rv,cc);
   }
   if (rv >= 0) {
      rv = cc;
      switch(cc) {
         case 0x00: 
	       if (rlen != 12) {
	         printf("Unexpected SOL response data received, len=%d\n",rlen);
	         rv = -1;
               }
               break;
         case 0x80: 
	       printf("SOL payload already active on another session\n");
               break;
         case 0x81: 
	       printf("SOL payload disabled\n");
               break;
         case 0x82: 
	       printf("SOL payload activation limit reached\n");
               break;
         case 0x83: 
	       printf("Cannot activate SOL payload with encryption\n");
               break;
         case 0x84: 
	       printf("Cannot activate SOL payload without encryption\n");
               break;
         default:
	       printf("Cannot activate SOL, ccode = 0x%02x\n",cc);
               break;
      }
   }
   if (rv == 0) {   /* success, use the response data */
      /* only here if response data is from IPMI 2.0 method */
      if (fdebug) dump_buf("sol act_resp",rbuf,rlen,0);
      in_payload_size  = rbuf[4] + (rbuf[5] << 8);
      out_payload_size = rbuf[6] + (rbuf[7] << 8);
      udp_port         = rbuf[8] + (rbuf[9] << 8);
      if (fdebug) printf("activate ok, in=%d out=%d port=%d\n",
			in_payload_size, out_payload_size, udp_port);
      if (bSolVer == 2) 
         lan2_set_sol_data(in_payload_size, out_payload_size, udp_port,
			   (void *)sol_output_handler);
   }
   return(rv);
}

int send_deactivate_sol(uchar verSOL)
{
   int rv = 0;
   uchar deactcmd;
   uchar netfn;
   uchar ibuf[64];
   uchar rbuf[64];
   int ilen, rlen;
   uchar cc;

   if (verSOL == 1) {
      return(send_deactivate_15sol());
   } else {   /* use IPMI 2.0 method */
      deactcmd  = DEACTIVATE_PAYLOAD;
      netfn   = NETFN_APP;
      ibuf[0] = PAYLOAD_TYPE_SOL;   /* payload type */
      ibuf[1] = 1;                  /* payload instance */
      ibuf[2] = 0;
      ibuf[3] = 0;
      ibuf[4] = 0;
      ibuf[5] = 0;
      ilen = 6;
   }
   rlen = sizeof(rbuf);
   rv = ipmi_cmdraw(deactcmd,netfn,BMC_SA, PUBLIC_BUS, BMC_LUN,
                   ibuf,(uchar)ilen,rbuf,&rlen,&cc,fdebug);
   if (fdebug) printf("sol deactivate rv = %d, cc = %x\n",rv,cc);
   if (rv == 0 && cc != 0) rv = cc;
   return(rv);
}


int sol_data_loop( void )
{
   int istdin;
   int rv = 0;
   uchar ibuf[0x64];
   int szibuf;
   int fd;
   int len;
   fd_set read_fds;
#ifdef WIN32
   HANDLE thnd;
#endif

   /* 
    * OK, now loop and listen to 
    *   stdin for user input 
    *   interface fd for incoming SOL data 
    */
   szibuf = sizeof(ibuf);  /* or via intf sol data */
   istdin = fileno(stdin);
   fd = sol_get_fd();
   if (fdebug) printf("stdin = %d, intf->fd = %d\n",istdin,fd);
   printf("[SOL session is running, use '%s' to end session.]\n",sol_esc_seq);

   tty_setraw(2);
#ifdef WIN32
   thnd = CreateThread(NULL, 0, &input_thread, NULL, 0, NULL);
   if (thnd == NULL) { 
       printf("CreateThread error, aborting\n");
       rv = -1; 
       done = 1; 
   }
   console_open();
#endif
   if (bSolVer == 1) {   /* jump start the session v1.5 data */
      ibuf[0] = 0x1b; /*escape*/
      ibuf[1] = '{';
      rv = sol_input_handler(ibuf,2);
   }

   while (done == 0) 
   {
       if (bKeepAlive > 0)   /* send KeepAlive */
           rv = sol_keepalive(bKeepAlive);   

       rv = os_select(istdin,fd, &read_fds);
       if (rv < 0) { 
	   if (fdebug) printf("os_select(%d) error %d\n",fd,rv); 
	   perror("select"); 
           done = 1; 
       } else if (rv > 0) {
           if (fdebug) printf("select rv = %d\n",rv);

           if (FD_ISSET(istdin, &read_fds)) {
               /* input from stdin (user) */
               memset(ibuf,0,szibuf);
               len = os_read(istdin, ibuf, szibuf);
               if (fdebug) printf("input: read len=%d, %02x %02x\n",
					len,ibuf[0], ibuf[1]);
               if (len <= 0) {
                  printf("Error %d reading stdin\n",len);
                  done = 1;
               } else {
                  rv = sol_input_handler(ibuf,len);
                  if (fdebug) printf("input: handler rv = %d\n",rv);
                  if (rv < 0) {
                      done = 1;
                      if (rv == RV_END) 
			   printf("%s exit via user input\n",progname);
                      else printf("Error %d processing input\n",rv);
                  }
               }
           }  /*endif stdin*/

           if (FD_ISSET(fd, &read_fds)) {
               SOL_RSP_PKT rs;
               rs.len = 0;
               /* output from BMC SOL */
               rv = sol_recv(&rs);
#ifdef WIN32
		/* Windows sometimes gets rv = -1 here */
               if (rv == -1) rv = -(WSAGetLastError());
#endif
               if (fdebug) printf("output: recv rv = %d\n",rv);
               if (rv < 0) {
                   printf("Error %d reading SOL socket\n",rv);
                   done = 1;
               } else sol_output_handler(&rs);
           }  /*endif fd*/
       }  /*endif select rv > 0 */
   }  /*end while*/
   if (rv == RV_END) rv = 0;
#ifdef WIN32
   CloseHandle(thnd);
   console_close();
#endif
   tty_setnormal(2);
   return(rv);
}

static void show_usage(void)
{
                printf("Usage: %s [-adcex -NUPREFTVY]\n", progname);
                printf(" where -a     activate SOL console session\n");
                printf("       -d     deactivate SOL console session\n");
                printf("       -c'^'  set escape Char to '^', default='~'\n");
                printf("       -e     Encryption off for SOL session\n");
                printf("       -l     Legacy mode for BIOS/DOS CR+LF\n");
#ifdef WIN32
                printf("       -r     Raw termio, no VT-ANSI translation\n");
                printf("       -w     use Windows Console buffer, not stdio\n");
#endif
                printf("       -x     show eXtra debug messages\n");
		print_lan_opt_usage();
}

#ifdef METACOMMAND
int i_sol(int argc, char **argv)
#else
#ifdef WIN32
int __cdecl
#else
int
#endif
main(int argc, char **argv)
#endif
{
   int ret = 0;
   int i;
   uchar devrec[16];
   uchar done = 0;
   char c;

   printf("%s ver %s\n", progname,progver);

   while ( (c = getopt( argc, argv,"ac:delrwxEF:T:V:N:P:R:U:Y?")) != EOF ) 
      switch(c) {
          case 'a': factivate = 1;   break;    /*activate*/
          case 'd': fdeactivate = 1; break;    /*deactivate*/
          case 'c':     		       /* escape char */
                if (strncmp(optarg,"0x",2) == 0) 
			sol_esc_seq[0] = htoi(&optarg[2]);  /*hex char value*/
                else if (optarg[0] >= '0' && optarg[0] <= '9') 
			sol_esc_seq[0] = atoi(optarg);  /*decimal char value*/
		else    sol_esc_seq[0] = optarg[0];  /*single ascii char*/
		break;
          case 'e': fSolEncryption = 0; break; /*encryption off*/
          case 'l': fCRLF = 1; break;          /*do legacy CR/LF translation*/
          case 'r': fRaw = 1; fCRLF = 0; break; /*raw termio, xlate off*/
          case 'w': fUseWinCon = 1;  break;    /*use Console routines*/
          case 'x': fdebug = 1;      break;    /*debug messages */
          case 'N':    /* nodename */
          case 'U':    /* remote username */
          case 'P':    /* remote password */
          case 'R':    /* remote password */
          case 'E':    /* get password from IPMI_PASSWORD environment var */
          case 'F':    /* force driver type */
          case 'T':    /* auth type */
          case 'V':    /* priv level */
          case 'Y':    /* prompt for remote password */
                parse_lan_options(c,optarg,fdebug);
                break;
	  default:
                show_usage();
                exit(1);
      }

   if (fipmi_lan == 0) {  /* no node specified */
      printf("Serial-Over-Lan Console requires a -N nodename\n");
      exit(1);
   }

   if (factivate == 0 && fdeactivate == 0) {
      show_usage();
      printf("Error: Need either -a or -d for sol\n");
      exit(1);
   }
#ifdef WIN32
   else if (factivate) {
      system("cls");
      printf("%s ver %s\n", progname,progver);
   }
#endif

   i = get_driver_type(); /*see if user explictly set type*/
   if (i == 0) bDriver = 0;  /*no driver type specified*/
   else bDriver = i;

   ret = ipmi_getdeviceid(devrec,16,fdebug);
   if (ret != 0) {
        show_outcome(progname,ret);  
	ipmi_close_();
	exit(1);
   } else {
      ipmi_maj = devrec[4] & 0x0f;
      ipmi_min = devrec[4] >> 4;
      mfg  = devrec[6] + (devrec[7] << 8) + (devrec[8] << 16); 
      printf("-- BMC version %x.%x, IPMI version %d.%d \n",
             devrec[2],  devrec[3], ipmi_maj, ipmi_min);
   }

   ret = ipmi_getpicmg(devrec,16,fdebug);
   if (ret == 0) fpicmg = 1;

   if (fdebug) printf("driver=%d mfg=%x\n",bDriver,mfg); 
   /* check for SOL support */
   if (ipmi_maj >= 2) { 
       if ((bDriver == DRV_LAN) && (mfg == VENDOR_INTEL)) {
          /* user specified to use IPMI LAN 1.5 SOL on Intel */
          bSolVer = 1;      /*IPMI 1.5 SOL*/
       } else {
          bSolVer = 2;      /*IPMI 2.0 SOL*/
          if (get_driver_type() == DRV_LAN) { /*now using IPMI LAN 1.5*/
            ipmi_close_();           /*close current IPMI LAN 1.5*/
            i = set_driver_type("lan2"); /*switch to IPMI LAN 2.0*/
          }
       } 
   } else if (ipmi_maj == 1) {
       if (ipmi_min >= 5) bSolVer = 1;  /* IPMI 1.5 */
       else bSolVer = 0;  /* IPMI 1.0 */
   } else bSolVer = 0;
   if (bSolVer == 0) {
       printf("Serial Over Lan not supported for this IPMI version\n");
       exit(1);
   }
#ifndef HAVE_LANPLUS
   if (bSolVer == 2) {
       printf("2.0 LanPlus module not available, trying 1.5 SOL instead\n");
       bSolVer = 1;
   }
#endif
   /* May also want to verify that SOL is implemented here */

   /*
    * Spawn a console raw terminal thread now, which will wait for the 
    * "Activating cmd (0x02)" on success 
    * Using globals: gnode,guser,gpswd, gauth_type, gpriv_level
    */
   if (fdeactivate) {
      ret = send_deactivate_sol(bSolVer); 
      if (fdebug) printf("send_deactivate_sol rv = %d\n",ret);
   } else if (factivate) {
      ret = send_activate_sol(bSolVer);
      if (fdebug) printf("send_activate_sol rv = %d\n",ret);
      if (ret == 0) {
         ret = sol_data_loop();
      } /*endif activate ok*/
   }  /*endif do activate */

   ipmi_close_();
   show_outcome(progname,ret); 
   exit (ret);
}  /* end main()*/

/* end isolconsole.c */
